{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Definition of an interval differentiation arithmetic which allows }
{    function evaluation with automatic differentiation up to second order.  }
{ Method: Overloading of operators and elementary functions for operations   }
{    of data type 'DerivType'.                                               }
{ Global types, operators, functions, and procedures:                        }
{    type       DerivType         : data type for differentiation arithmetic }
{    operators  +, -, *, /        : operators of differentiation arithmetic  }
{    functions  DerivConst,                                                  }
{               DerivVar          : to define derivative constants/variables }
{    functions  fValue,                                                      }
{               dfValue,                                                     }
{               ddfValue          : to get function and derivative values    }
{    functions  sqr, sqrt, power,                                            }
{               exp, sin, cos,... : elementary functions of diff. arithmetic }
{    procedure  fEval(...)        : to compute function value only           }
{    procedure  dfEval(...)       : to compute function and first derivative }
{                                   value                                    }
{    procedure  ddfEval(...)      : to compute function, first, and second   }
{                                   derivative value                         }
{----------------------------------------------------------------------------}
module ddf_ari;

use i_ari, i_util;         { interval arithmetic, interval utility functions }

{----------------------------------------------------------------------------}
{ Global type definition and local variable                                  }
{----------------------------------------------------------------------------}
global type
  DerivType      = record f, df, ddf : interval; end;

var                  { The local variable 'DerivOrder' is used to select the }
  DerivOrder : 0..2; { highest order of derivative which is computed. Its    }
                     { default value is 2, and normally the first and the    }
                     { second derivatives are computed.                      }
{----------------------------------------------------------------------------}
{ Transfer functions for constants and variables                             }
{----------------------------------------------------------------------------}
global function DerivConst (c: real) : DerivType;        { Generate constant }
begin                                                    {-------------------}
  DerivConst.f := c;  DerivConst.df := 0;  DerivConst.ddf := 0;
end;

global function DerivConst (c: interval) : DerivType;    { Generate constant }
begin                                                    {-------------------}
  DerivConst.f := c;  DerivConst.df := 0;  DerivConst.ddf := 0;
end;

global function DerivVar (v: real) : DerivType;          { Generate variable }
begin                                                    {-------------------}
  DerivVar.f := v;  DerivVar.df := 1;  DerivVar.ddf := 0;
end;

global function DerivVar (v: interval) : DerivType;      { Generate variable }
begin                                                    {-------------------}
  DerivVar.f := v;  DerivVar.df := 1;  DerivVar.ddf := 0;
end;

{----------------------------------------------------------------------------}
{ Access functions for function and derivative values                        }
{----------------------------------------------------------------------------}
global function fValue (u: DerivType) : interval;       { Get function value }
begin                                                   {--------------------}
  fValue:= u.f;
end;

global function dfValue (u: DerivType) : interval; { Get 1. derivative value }
begin                                              {-------------------------}
  dfValue:= u.df;
end;

global function ddfValue (u: DerivType) : interval;{ Get 2. derivative value }
begin                                              {-------------------------}
  ddfValue:= u.ddf;
end;

{----------------------------------------------------------------------------}
{ Monadic operators + and - for DerivType operands                           }
{----------------------------------------------------------------------------}
global operator + (u: DerivType) res: DerivType;
begin
  res:= u;
end;

global operator - (u: DerivType) res: DerivType;
begin
  res.f := -u.f;
  if (DerivOrder > 0) then
  begin
    res.df := -u.df;
    if (DerivOrder > 1) then res.ddf:= -u.ddf;
  end;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for two DerivType operands                        }
{----------------------------------------------------------------------------}
global operator + (u,v: DerivType) res: DerivType;
begin
  res.f := u.f + v.f;
  if (DerivOrder > 0) then
  begin
    res.df := u.df + v.df;
    if (DerivOrder > 1) then res.ddf:= u.ddf + v.ddf;
  end;
end;

global operator - (u,v: DerivType) res: DerivType;
begin
  res.f := u.f   - v.f;
  if (DerivOrder > 0) then
  begin
    res.df := u.df - v.df;
    if (DerivOrder > 1) then res.ddf:= u.ddf - v.ddf;
  end;
end;

global operator * (u,v: DerivType) res: DerivType;
begin
  res.f := u.f*v.f;
  if (DerivOrder > 0) then
  begin
    res.df := u.df*v.f + u.f*v.df;
    if (DerivOrder > 1) then res.ddf:= u.ddf*v.f + 2*u.df*v.df + u.f*v.ddf;
  end;
end;

global operator / (u,v: DerivType) res: DerivType;
var h1, h2: interval;
begin
  h1 := u.f/v.f;  { Can propagate 'division by zero' error }
  res.f := h1;
  if (DerivOrder > 0) then
  begin
    h2 := (u.df - h1*v.df)/v.f;  res.df := h2;
    if (DerivOrder > 1) then res.ddf:= (u.ddf - h1*v.ddf - 2*h2*v.df)/v.f;
  end;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for one interval and one DerivType operand        }
{----------------------------------------------------------------------------}
global operator + (u: interval; v: DerivType) res: DerivType;
begin
  res.f := u + v.f;
  if (DerivOrder > 0) then
  begin
    res.df := v.df;
    if (DerivOrder > 1) then res.ddf:= v.ddf;
  end;
end;

global operator - (u: interval; v: DerivType) res: DerivType;
begin
  res.f := u - v.f;
  if (DerivOrder > 0) then
  begin
    res.df := - v.df;
    if (DerivOrder > 1) then res.ddf:= - v.ddf;
  end;
end;

global operator * (u: interval; v: DerivType) res: DerivType;
begin
  res.f := u*v.f;
  if (DerivOrder > 0) then
  begin
    res.df := u*v.df;
    if (DerivOrder > 1) then res.ddf:= u*v.ddf;
  end;
end;

global operator / (u: interval; v: DerivType) res: DerivType;
var h1, h2: interval;
begin
  h1 := u/v.f;  { Can propagate 'division by zero' error }
  res.f := h1;
  if (DerivOrder > 0) then
  begin
    h2 := -h1*v.df/v.f;  res.df := h2;
    if (DerivOrder > 1) then res.ddf:= (-h1*v.ddf - 2*h2*v.df)/v.f;
  end;
end;

global operator + (u: DerivType; v: interval) res: DerivType;
begin
  res.f := u.f + v;
  if (DerivOrder > 0) then
  begin
    res.df := u.df;
    if (DerivOrder > 1) then res.ddf:= u.ddf;
  end;
end;

global operator - (u: DerivType; v: interval) res: DerivType;
begin
  res.f := u.f - v;
  if (DerivOrder > 0) then
  begin
    res.df := u.df;
    if (DerivOrder > 1) then res.ddf:= u.ddf;
  end;
end;

global operator * (u: DerivType; v: interval) res: DerivType;
begin
  res.f := u.f * v;
  if (DerivOrder > 0) then
  begin
    res.df := u.df * v;
    if (DerivOrder > 1) then res.ddf:= u.ddf * v;
  end;
end;

global operator / (u: DerivType; v: interval) res: DerivType;
begin
  res.f := u.f / v;  { Can propagate 'division by zero' error }
  if (DerivOrder > 0) then
  begin
    res.df := u.df / v;
    if (DerivOrder > 1) then res.ddf:= u.ddf / v;
  end;
end;

{----------------------------------------------------------------------------}
{ Operators +, -, *, and / for one real and one DerivType operand            }
{----------------------------------------------------------------------------}
global operator + (u: real; v: DerivType) res: DerivType;
begin
  res := intval(u) + v;
end;

global operator - (u: real; v: DerivType) res: DerivType;
begin
  res := intval(u) - v;
end;

global operator * (u: real; v: DerivType) res: DerivType;
begin
  res := intval(u) * v;
end;

global operator / (u: real; v: DerivType) res: DerivType;
begin
  res := intval(u) / v;  { Can propagate 'division by zero' error }
end;

global operator + (u: DerivType; v: real) res: DerivType;
begin
  res := u + intval(v);
end;

global operator - (u: DerivType; v: real) res: DerivType;
begin
  res := u - intval(v);
end;

global operator * (u: DerivType; v: real) res: DerivType;
begin
  res := u * intval(v);
end;

global operator / (u: DerivType; v: real) res: DerivType;
begin
  res := u / intval(v);  { Can propagate 'division by zero' error }
end;

{----------------------------------------------------------------------------}
{ Elementary functions for DerivType arguments                               }
{----------------------------------------------------------------------------}
global function sqr (u: DerivType) : DerivType;
begin
  sqr.f  := sqr(u.f);
  if (DerivOrder > 0) then
  begin
    sqr.df := 2*u.f*u.df;
    if (DerivOrder > 1) then sqr.ddf:= 2 * (sqr(u.df) + u.f*u.ddf);
  end;
end;

global function power (u: DerivType; k: integer) : DerivType;
var
  h1 : interval;
begin
  if (k = 0) then
    power:= DerivConst(1)
  else if (k = 1) then
    power:= u
  else if (k = 2) then
    power:= sqr(u)
  else
    begin
      power.f:= power(u.f, k);
      if (DerivOrder > 0) then
      begin
        h1 := k * power(u.f, k-1);  power.df:= h1 * u.df;
        if (DerivOrder > 1) then
          power.ddf:= h1 * u.ddf + k*(k-1)*power(u.f, k-2)*sqr(u.df);
      end;
    end;
end;

global function sqrt (u: DerivType) : DerivType;
var h1, h2: interval;
begin
  h1 := sqrt(u.f);  { Can propagate domain error }
  sqrt.f := h1;
  if (DerivOrder > 0) then
  begin
    h1 := 0.5/h1;  h2 := u.df*h1;  sqrt.df := h2;
    if (DerivOrder > 1) then sqrt.ddf:= u.ddf*h1 - 0.5*u.df/u.f*h2;
  end;
end;

global function exp (u: DerivType) : DerivType;
var h1: interval;
begin
  h1 := exp(u.f);  exp.f := h1;
  if (DerivOrder > 0) then
  begin
    exp.df := h1*u.df;
    if (DerivOrder > 1) then exp.ddf:= (u.ddf + sqr(u.df))*h1;
  end;
end;

global function ln (u: DerivType) : DerivType;
var h: interval;
begin
  ln.f := ln(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h := u.df/u.f;   ln.df := h;
    if (DerivOrder > 1) then ln.ddf:= (u.ddf - h*u.df) / u.f;
  end;
end;

global function sin (u: DerivType) : DerivType;
var h0, h1: interval;
begin
  h0 := sin(u.f);  sin.f := h0;
  if (DerivOrder > 0) then
  begin
    h1 := cos(u.f);  sin.df := h1*u.df;
    if (DerivOrder > 1) then sin.ddf:= h1*u.ddf - h0*sqr(u.df);
  end;
end;

global function cos (u: DerivType) : DerivType;
var h0, h1: interval;
begin
  h0 := cos(u.f);  cos.f := h0;
  if (DerivOrder > 0) then
  begin
    h1 := -sin(u.f);  cos.df := h1*u.df;
    if (DerivOrder > 1) then cos.ddf:= h1*u.ddf - h0*sqr(u.df);
  end;
end;

global function tan (u: DerivType) : DerivType;
var h0, h1, h2: interval;
begin
  h0 := tan(u.f);  { Can propagate domain error }
  tan.f := h0;
  if (DerivOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h1 := sqr(h0)+1;  h2 := 2*h0*h1;      {   h0 * (h0^2 + 1) <= h0^3 + h0   }
    tan.df := h1*u.df;                    { So, we use the first form.       }
    if (DerivOrder > 1) then tan.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function cot (u: DerivType) : DerivType;
var h0, h1, h2: interval;
begin
  h0 := cot(u.f);  { Can propagate domain error }
  cot.f := h0;
  if (DerivOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h1 := -(sqr(h0)+1);  h2 := -2*h0*h1;  {   h0 * (h0^2 + 1) <= h0^3 + h0   }
    cot.df := h1*u.df;                    { So, we use the first form.       }
    if (DerivOrder > 1) then cot.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function arcsin (u: DerivType) : DerivType;
var h, h1, h2: interval;
begin
  arcsin.f := arcsin(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h:= 1 - sqr(u.f);  h1:= 1/sqrt(h);  arcsin.df:= h1*u.df;  h2:= u.f*h1/h;
    if (DerivOrder > 1) then arcsin.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function arccos (u: DerivType) : DerivType;
var h, h1, h2: interval;
begin
  arccos.f := arccos(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h:= 1 - sqr(u.f);  h1:= -1/sqrt(h);  arccos.df:= h1*u.df;  h2:= u.f*h1/h;
    if (DerivOrder > 1) then arccos.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function arctan (u: DerivType) : DerivType;
var h1,h2: interval;
begin
  arctan.f := arctan(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h1 := 1 / (1 + sqr(u.f));  arctan.df := h1*u.df;  h2 := -2*u.f*sqr(h1);
    if (DerivOrder > 1) then arctan.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function arccot (u: DerivType) : DerivType;
var h1, h2: interval;
begin
  arccot.f := arccot(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h1 := -1 / (1 + sqr(u.f));  arccot.df := h1*u.df;  h2 := 2*u.f*sqr(h1);
    if (DerivOrder > 1) then arccot.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function sinh (u: DerivType) : DerivType;
var h0, h1: interval;
begin
  h0 := sinh(u.f);  sinh.f := h0;
  if (DerivOrder > 0) then
  begin
    h1 := cosh(u.f);  sinh.df := h1*u.df;
    if (DerivOrder > 1) then sinh.ddf:= h1*u.ddf + h0*sqr(u.df);
  end;
end;

global function cosh (u: DerivType) : DerivType;
var h0, h1: interval;
begin
  h0 := cosh(u.f);  cosh.f := h0;
  if (DerivOrder > 0) then
  begin
    h1 := sinh(u.f);  cosh.df := h1*u.df;
    if (DerivOrder > 1) then cosh.ddf:= h1*u.ddf + h0*sqr(u.df);
  end;
end;

global function tanh (u: DerivType) : DerivType;
var h0, h1, h2: interval;
begin
  h0 := tanh(u.f);  tanh.f := h0;
  if (DerivOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h1 := 1 - sqr(h0);  h2 := -2*h0*h1;   {   h0 * (h0^2 - 1) <= h0^3 - h0   }
    tanh.df := h1*u.df;                   { So, we use the first form.       }
    if (DerivOrder > 1) then tanh.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function coth (u: DerivType) : DerivType;
var h0, h1, h2: interval;
begin
  h0 := coth(u.f);  { Can propagate domain error }
  coth.f := h0;
  if (DerivOrder > 0) then
  begin                                   { The subdistributive law implies  }
    h1 := 1 - sqr(h0);  h2 := -2*h0*h1;   {   h0 * (h0^2 - 1) <= h0^3 - h0   }
    coth.df := h1*u.df;                   { So, we use the first form.       }
    if (DerivOrder > 1) then coth.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function arsinh (u: DerivType) : DerivType;
var h, h1, h2: interval;
begin
  arsinh.f := arsinh(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h:= 1 + sqr(u.f);  h1:= 1/sqrt(h);  arsinh.df:= h1*u.df;  h2:= -u.f*h1/h;
    if (DerivOrder > 1) then arsinh.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function arcosh (u: DerivType) : DerivType;
var h, h1, h2: interval;
begin
  arcosh.f := arcosh(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h:= sqr(u.f) - 1;  h1:= 1/sqrt(h);  arcosh.df:= h1*u.df;  h2:= -u.f*h1/h;
    if (DerivOrder > 1) then arcosh.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function artanh (u: DerivType) : DerivType;
var h1,h2: interval;
begin
  artanh.f := artanh(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h1 := 1 / (1 - sqr(u.f));  artanh.df := h1*u.df;  h2 := 2*u.f*sqr(h1);
    if (DerivOrder > 1) then artanh.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

global function arcoth (u: DerivType) : DerivType;
var h1, h2: interval;
begin
  arcoth.f := arcoth(u.f);  { Can propagate domain error }
  if (DerivOrder > 0) then
  begin
    h1 := 1 / (1 - sqr(u.f));  arcoth.df := h1*u.df;  h2 := 2*u.f*sqr(h1);
    if (DerivOrder > 1) then arcoth.ddf:= h1*u.ddf + h2*sqr(u.df);
  end;
end;

{----------------------------------------------------------------------------}
{ Predefined routines for evaluation of DerivType-functions                  }
{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing only the function value.                           }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'DerivType'.                           }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{ Description: This procedure sets 'DerivOrder' to 0, evaluates 'f(x)' in    }
{    differentiation arithmetic, and returns the function value only.        }
{----------------------------------------------------------------------------}
global procedure fEval (function f(x:DerivType) : DerivType;
                                              x : interval;
                             var             fx : interval);
var
  fxD : DerivType;
begin
  DerivOrder := 0;  fxD := f(DerivVar(x));  fx := fxD.f;  DerivOrder := 2;
end;

{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing the function value and the value of the first      }
{    derivative.                                                             }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'HessType'.                            }
{             'x'       : argument for evaluation of 'f'.                    }
{    Out    : 'fx'      : returns the function value 'f(x)'.                 }
{             'dfx'     : returns the first derivative value 'f'(x)'.        }
{ Description: This procedure sets 'DerivOrder' to 1, evaluates 'f(x)' in    }
{    differentiation arithmetic, and returns the function value and the      }
{    value of the first derivative.                                          }
{----------------------------------------------------------------------------}
global procedure dfEval (function f(x:DerivType) : DerivType;
                                               x : interval;
                              var        fx, dfx : interval);
var
  fxD : DerivType;
begin
  DerivOrder:= 1;  fxD:= f(DerivVar(x));  fx:= fxD.f;  dfx:= fxD.df;
  DerivOrder:= 2;
end;

{----------------------------------------------------------------------------}
{ Purpose: Evaluation of function 'f' for argument 'x' in differentiation    }
{    arithmetic computing the function value, the value of the first, and    }
{    the value of the second derivative.                                     }
{ Parameters:                                                                }
{    In     : 'f'     : function of 'HessType'.                              }
{             'x'     : argument for evaluation of 'f'.                      }
{    Out    : 'fx'    : returns the function value 'f(x)'.                   }
{             'dfx'   : returns the value of the first derivative 'f'(x)'.   }
{             'ddfx'  : returns the value of the second derivative 'f''(x)'. }
{ Description: This procedure keeps 'DerivOrder' = 2, evaluates 'f(x)' in    }
{    differentiation arithmetic, and returns the function value, the value   }
{    of the first, and the value of the second derivative.                   }
{----------------------------------------------------------------------------}
global procedure ddfEval (function f(x:DerivType) : DerivType;
                                                x : interval;
                               var  fx, dfx, ddfx : interval);
var
  fxD : DerivType;
begin
  fxD:= f(DerivVar(x));  fx:= fxD.f;  dfx:= fxD.df;  ddfx:= fxD.ddf;
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  DerivOrder := 2;
end.
