{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Determine enclosures for the optimal value 'z_opt', for the set   }
{    of optimal basic index sets 'V_opt', and for the set of solution        }
{    vectors 'X_opt' for a linear programming problem P = (A,b,c) given in   }
{    the standard form:                                                      }
{               ( z = c^t * x = max! )                                       }
{         (LP)  (     A * x = b      )                                       }
{               (      x >= 0        )                                       }
{    with an initial optimal basic index set.                                }
{ Method: Starting from an initial basic index set, all neighboring index    }
{    sets are examined to determine all optimal solutions. To validate the   }
{    optimality of a solution, all linear systems of equations are solved    }
{    using a verifying linear system solver.                                 }
{ Global procedures and functions:                                           }
{   procedure LinOpt(...)      : determines the enclosures of z_opt, V_opt,  }
{                                and X_opt for LP P = (A,b,c)                }
{   function LinOptErrMsg(...) : delivers an error message text              }
{ Updates:                                                                   }   
{   11.7.94 : Splitting of calculation  NewBase in 'NeighboringList'         }
{----------------------------------------------------------------------------}
module lop; { Linear optimization module }
            { -      --                  }

use     i_ari,           { Interval arithmetic               }
        mv_ari,          { Matrix/vector arithmetic          }
        mvi_ari,         { Matrix/vector interval arithmetic }
        linsys,          { Verifying linear system solver    }
        x_real,          { Extended real arithmetic          }
        lop_ari;         { Linear linked list for index sets }

const  kmax = 100;       { Maximum number of iterations      }
var    Rplus   : interval;
{----------------------------------------------------------------------------}
{ Error codes used in this module.                                           }
{----------------------------------------------------------------------------}
const
  NoError                 = 0;  { No error occurred                          }
  WrongDimension          = 1;  { Wrong dimension of problem                 }
  SolutionSetIsEmpty      = 2;  { Set of feasible solutions is empty         }
  FunctionUnbounded       = 3;  { Objective function is unbounded            }
  SubmatrixSingular       = 4;  { Submatrix A_B is singular                  }
  StartIndexSetNotOptimal = 5;  { Initial basic index set not optimal        }
  IterationError          = 6;  { Maximum number of iteration exceeded       }
  SolutionMatrixTooSmall  = 7;  { Matrices for storage of solutions too small}
  NoOptimalSolution       = 8;  { No initial optimal solution found          }
  NoInitialVertex         = 9;  { No initial vertex found                    }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function LinOptErrMsg(Err : integer) : string;
var
  Msg : string;
begin
  case Err of
    NoError           : Msg := '';
    WrongDimension    : Msg := 'Wrong dimension of problem (e.g. m >= n)';
    SolutionSetIsEmpty: Msg := 'Set of feasible solutions is probably empty';
    FunctionUnbounded : Msg := 'Objective function is probably unbounded';
    SubmatrixSingular : Msg := 'Submatrix A_B in [A_B] is probably singular';
    StartIndexSetNotOptimal   
                      : Msg := 'Initial basic index set not optimal '
                               + '(i.e. B_start not in V_opt)';
    IterationError    : Msg := 'Maximum number of iterations (='
                               + image(kmax,0) + ') exceeded';
    SolutionMatrixTooSmall
                      : Msg := 'Matrices for storage of solutions is'
                               + ' too small. Increase index ranges of V_opt'
                               + ' and X_opt';
    NoOptimalSolution : Msg := 'No initial optimal solution found';
    NoInitialVertex   : Msg := 'No initial vertex found';
    else              : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  LinOptErrMsg := Msg;
end;
                                                 { Determine solution vector }
function Solution(TT            : imatrix;       { corresponding to Base     }
                  Base, NonBase : IndexSet)      {---------------------------}
                  : ivector[1..size(Base)+size(NonBase)]; 
var Index, i : integer;
begin
  Index := 0;
  Solution := 0;
  for i:=1 to size(Base) do
    begin
      Index := NextIndex(Index,Base);
      Solution[Index] := TT[i+1,1] ** Rplus;
    end;
end;


{----------------------------------------------------------------------------}
{ Compute the interval tableau                                               }
{                            ([z]   | [d]^t    )                             }
{                     [T] := (------+----------)                             }
{                            ([x]_B | [H]      ).                            }
{----------------------------------------------------------------------------}
procedure ComputeTableau(       A             : rmatrix;
                                b, c          : rvector;
                                Base, NonBase : IndexSet;
                            var TT            : imatrix;
                            var Err           : integer);
var
  A_B                : rmatrix[1..ub(A,1), 1..size(Base)];
  A_N                : rmatrix[1..ub(A,1), 1..size(NonBase)];
  c_B                : rvector[1..size(Base)];
  c_N                : rvector[1..size(NonBase)];
  xx_B, yy           : ivector[1..size(Base)];
  dd                 : ivector[1..size(NonBase)];
  HH                 : imatrix[1..size(Base), 1..size(NonBase)];
  zz                 : interval;
  Index, local_Error : integer;
  i, j, m, n         : integer;

begin { procedure ComputeTableau }

  m := ub(A,1);
  n := ub(A,2);

  { Determine submatrices and subvectors according to index sets Base and    }
  { NonBase                                                                  }
  A_B := extract(A, Base); A_N := extract(A, NonBase);{ Determine submatrices}
  c_B := extract(c, Base); c_N := extract(c, NonBase);{ and subvectors accor-}
                                                      { ding to B and N      }
  { Solve linear systems of equations }               {----------------------}
  linsolve(A_B, b, xx_B, local_Error);

  if (local_Error = NoError) then  
    linsolve(transp(A_B), c_B, yy, local_Error);

  Index := 0;
  if (local_Error = NoError) then  
    for i:=1 to (n-m) do
      begin 
        Index := NextIndex(Index,NonBase);
        if (local_Error=NoError) then  
        linsolve(A_B, A[*,Index], HH[*,i], local_Error);
      end;

  if (local_Error = NoError) then
    begin 
      dd      := transp(A_N) * yy - c_N;    { Compute components of interval }
      zz      := (c_B * xx_B) ** (b * yy);  { tableau                        }
      TT[1,1] := zz;                        {--------------------------------}
      for j:=2 to (n-m+1) do
        TT[1,j] := dd[j-1];
      for i:=2 to (m+1) do
        TT[i,1] := xx_B[i-1];
      for i:=2 to (m+1) do
        for j:=2 to (n-m+1) do
          TT[i,j] := HH[i-1,j-1];
    end
  else Err := SubmatrixSingular;

end; { procedure ComputeTableau }

{----------------------------------------------------------------------------}
{ Determine whether the interval tableau [T] represents a basis stable       }
{ solution, i.e. ([x]_B.inf > 0) and ([d].inf > 0).                          }
{----------------------------------------------------------------------------}
function BasisStable(TT : imatrix) : boolean;
var IsStable : boolean;
    i,j      : integer;
begin
  IsStable := true; i := 1; j := 1;
  while ( (i < ub(TT,1)) and IsStable ) do
    begin
      i := i+1; { [x]_B = [T_i,1] (i = 2 .. m+1) }
      if (TT[i,1].inf <= 0) then IsStable := false;
    end;
  while ( (j < ub(TT,2)) and IsStable ) do
    begin
      j := j+1; { [d] = [T_1,j] (j = 2 .. n-m+1) }
      if (TT[1,j].inf <= 0) then IsStable := false;
    end;
  BasisStable := IsStable;
end;

{----------------------------------------------------------------------------}
{ Determine whether the interval tableau [T] represents a possibly optimal   }
{ solution, i.e ([x]_B.sup >= 0) and ([d].sup >= 0)                          }
{----------------------------------------------------------------------------}
function PossiblyOptimalSolution(TT : imatrix) : boolean;
var IsOptimal : boolean;
    i,j       : integer;
begin
  IsOptimal := true; i := 1; j := 1;
  while ( (i < ub(TT,1)) and IsOptimal ) do
    begin
      i := i+1; { [x]_B = [T_i,1] (i = 2 .. m+1) }
      if not( TT[i,1].sup >= 0 ) then IsOptimal := false;
    end;
  while ( (j < ub(TT,2)) and IsOptimal ) do
    begin
      j := j+1; { [d] = [T_1,j] (j = 2 .. n-m+1) }
      if not( TT[1,j].sup >= 0 ) then IsOptimal := false;
    end;
  PossiblyOptimalSolution := IsOptimal;
end;




{----------------------------------------------------------------------------}
{ Check whether the set of feasible solutions is empty, i.e. not for all(beta}
{ in B) with (0 in [x]_beta) there exists a (nu in N) with ([H]_beta,nu < 0) }
{----------------------------------------------------------------------------}
function EmptySolutionSet(TT : imatrix) : boolean;
var IsEmpty : boolean;
    i,j     : integer;
begin
  IsEmpty := false; i := 1;
  while (not IsEmpty) and (i < ub(TT,1)) do
    begin
      i := i+1;
      if (0 in TT[i,1]) then
        begin
          IsEmpty := true;
          j := 1;
          while IsEmpty and (j < ub(TT,2)) do
            begin
              j := j+1;
              if (TT[i,j].sup < 0) then IsEmpty := false;
            end;
        end;
    end;
  EmptySolutionSet := IsEmpty;
end;

{----------------------------------------------------------------------------}
{ Check whether the objective function is unbounded, i.e. not for all (nu    }
{ in N) with (0 in [d]_nu) there exists a (beta in B) with ([H]_beta,nu > 0) }
{----------------------------------------------------------------------------}
function Unbounded(TT : imatrix) : boolean;
var IsUnbounded : boolean;
    i,j         : integer;
begin
  IsUnbounded := false; j := 1;
  while (not IsUnbounded) and (j < ub(TT,2)) do
    begin
      j := j+1;
      if (0 in TT[1,j]) then
        begin
          IsUnbounded := true;
          i := 1;
          while IsUnbounded and (i < ub(TT,1)) do
            begin
              i := i+1;
              if (TT[i,j].inf > 0) then IsUnbounded := false;
            end;
        end;
    end;
  Unbounded := IsUnbounded;
end;

{----------------------------------------------------------------------------}
{ Determine list of neighboring basic index sets L for index set Base that   }
{ are good candidates for being optimal basic index sets.                    }
{----------------------------------------------------------------------------}
function NeighboringList(TT : imatrix; Base, NonBase : IndexSet) : BaseList;
var
  L                 : BaseList;
  NewBase           : IndexSet;
  i, j, m, n        : integer;
  beta, nu, Counter : integer;
  colmin, rowmax    : real;
  xx, dd, HH        : interval;

begin 
  m := size(Base); n := m + size(NonBase);                  { Initialization }
  L := nil;                                                 {----------------}

  colmin := maxReal;                           { Search for candidates (nu   }
  for j := 1 to (n-m) do { for all (nu in N) } { in N) and (beta in B) by    }
    begin                                      { determining of the minimum  }
      dd := TT[1,j+1];                         { of the quotients            }
      if (0 in dd) then                        { [x]_beta / [H]_beta,nu      }
        begin                                  {-----------------------------}
          for i:=1 to m do { for all (beta in B) } { Determine minimum of    }
            begin                                  {  [x]_beta / [H]_beta,nu }
              xx := TT[i+1,1];                     { for column nu of [T]    }
              HH := TT[i+1,j+1];                   {-------------------------}
              if ((HH.Inf > 0) and (xx.Sup/<HH.Inf < colmin)) then 
                colmin := xx.Sup/<HH.Inf;
            end; { (beta in B) }
          for i:=1 to m do { for all (beta in B) }{ Determine candidate (beta}
            begin                                 { in B) for exchange of    }
              xx := TT[i+1,1];                    { indices                  }
              HH := TT[i+1,j+1];                  {--------------------------}
              if (HH.Sup > 0) and (xx.Inf/<HH.Sup <= colmin) then 
                begin                          { Determine new index set and }
                  beta := GetIndex(i,Base);    { add it to L                 }
                  nu   := GetIndex(j,NonBase); {-----------------------------}
                  NewBase := Base - [beta];    {    Splitting    }
                  NewBase := NewBase + [nu];   { Changed 11.7.94 }
                  insert(L, NewBase);          {-----------------}
                end;
            end; { (beta in B) }
      end;
    end; { (nu in N) }

  rowmax := -maxReal;                          { Search for candidates (beta }
  for i := 1 to m do { for all (beta in B) }   { in B) and (nu in N) by      }
    begin                                      { determining of the maximum  }
      xx := TT[i+1,1];                         { of the quotients            }
      if (0 in xx) then                        { [d]_nu / [H]_beta,nu        }
        begin                                  {-----------------------------}
          for j:=1 to (n-m) do { for all (nu in N) } { Determine maximum of  }
            begin                                    { [d]_nu / [H]_beta,nu  }
              dd := TT[1,j+1];                       { for row beta of [T]   }
              HH := TT[i+1,j+1];                     {-----------------------}
              if ((HH.Sup < 0) and (dd.Inf/>HH.Sup > rowmax)) then 
                rowmax := dd.Sup/>HH.Sup;
            end; { (nu in N) }
          for j:=1 to (n-m) do { for all (nu in N) } { Determine candidate   }
            begin                                    { (nu in N) for exchange}
              dd := TT[1,j+1];                       { of indices            }
              HH := TT[i+1,j+1];                     {-----------------------}
              if (HH.Inf < 0) and (dd.Inf/>HH.Inf >= rowmax) then 
                begin                          { Determine new index set and }
                  beta := GetIndex(i,Base);    { add it to L                 }
                  nu := GetIndex(j,NonBase);   {-----------------------------}
                  NewBase := Base - [beta];    {    Splitting    }
                  NewBase := NewBase + [nu];   { Changed 11.7.94 }
                  insert(L, NewBase);          {-----------------}
                end;
            end; { (nu in N) }
      end;
    end; { (beta in B) }

  NeighboringList := L;

end;  { function NeighboringList }


{----------------------------------------------------------------------------}
{ Purpose: Determination of enclosures for the solutions of a linear opti-   }
{    mization problem:      ( z = c^t * x = max! )                           }
{                     (LP)  (     A * x = b      )                           }
{                           (      x >= 0        )                           }
{   with an initial optimal basic index set.                                 }
{ Parameters:                                                                }
{    In: 'A'    : matrix of constraints                                      }
{        'b'    : right-hand side of constraints                             }
{        'c'    : represents the objective function by its coefficients      }
{        'B_st.': initial basic index set                                    }
{    Out:'z_opt': enclosure of the optimal value                             }
{        'V_opt': superset of all optimal basic index sets                   }
{        'X_opt': superset of all optimal solution vectors                   }
{ Description: Starting from an initial basic index set 'B_start', a true    }
{    superset of all neighboring index sets 'L' that are candidates for being}
{    optimal index sets are determined and stored in a list of candidates    }
{    'CList'. As long as there are candidates left, they are examined using  }
{    a verifying linear system solver. The linear system solver returns      }
{    enclosures for the solutions of the square linear systems A_B [x] = b,  }
{    where A_B denotes a submatrix of the constraint matrix A according to a }
{    basic index set B.                                                      }
{----------------------------------------------------------------------------}
global procedure LinOpt  (      A                       : rmatrix;
                                b, c, B_start_Vector    : rvector;
                           var  z_opt                   : interval;
                           var  V_opt                   : rmatrix;
                           var  X_opt                   : imatrix;
                           var  No, Err                 : integer    );

var 
  B_start           : IndexSet;
  Base, NonBase     : IndexSet;
  L, CList, E       : BaseList;
  TT                : imatrix[1..1+(ub(b)), 1..1+(ub(c)-ub(b))];
  k, m, n, i, maxNo : integer;

begin { global procedure LinOpt }

  Err := NoError;

  m := ub(A,1); n := ub(A,2);               { Determine dimension of problem }
                                            {--------------------------------}
  if (n <= m) or (m < 1) or (n < 1) or (n > maxDim)       { Check dimensions }
              or (ub(V_opt,1) <> ub(X_opt,1)) then        {------------------}
    Err := WrongDimension
  else { Err <> WrongDimension }
    begin
      B_start := [];                                        { Initialization }
      for i:=1 to m do  B_start := B_start + [trunc(B_start_Vector[i])];
      k := 0; V_opt := 0; X_opt := 0; No := 0; maxNo := ub(V_opt,1);
      E := nil; CList := nil; insert(CList, B_start);

      repeat { until empty(CList) }

        Base := select(CList); NonBase := [1..n] - Base; { Select Base out of}
        delete(CList, Base); insert(E, Base);            { CList and update  }
                                                         { examined list E   }
        ComputeTableau(A, b, c, Base, NonBase, TT, Err); {-------------------}

        if (Err = NoError) then

          if BasisStable(TT) then
            begin
              No          := No + 1;                      { Store unique     }
              z_opt       := TT[1,1];                     { optimal solution }
              X_opt[No,*] := Solution(TT, Base, NonBase); {------------------}
              V_opt[No,*] := Base;
            end

          else { not basis stable }
            begin
              if EmptySolutionSet(TT) then Err := SolutionSetIsEmpty
              else if Unbounded(TT)   then Err := FunctionUnbounded

              else if PossiblyOptimalSolution(TT) and (Err = NoError) then
                begin
                  No          := No + 1;                     { Store optimal }
                  if (No = 1) then z_opt:= TT[1,1]           { solution      }
                              else z_opt:= TT[1,1] +* z_opt; {---------------}
                  X_opt[No,*] := Solution(TT, Base, NonBase);
                  V_opt[No,*] := Base;

                  L := NeighboringList(TT,Base,NonBase); { Determine list of }
                  remove(L, E); remove(L, CList);        { neighboring basic }
                                                         { index sets        }
                                                         {-------------------}
                  append(CList, L);         { Compute new list of candidates }
                end;                        {--------------------------------}
            end;
  
        k := k+1;

      until empty(CList) or (Err <> NoError) or (k = kmax) or (No = maxNo);

      if (Err = NoError) then                         { Determine error code }
        if (not empty(CList)) and (k = kmax) then     {----------------------}
                              Err := IterationError
        else if (No = 0) then Err := StartIndexSetNotOptimal
        else if (No = maxNo) and (not empty(CList)) then 
                              Err := SolutionMatrixTooSmall;

      FreeAll(CList); FreeAll(E);

    end; { Err <> WrongDimension }

end;  { global procedure LinOpt }

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  Rplus := intval(0,maxReal);
end.
