{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Computing enclosures for all zeros of a continuously              }
{    differentiable one-dimensional, real-valued function.                   }
{ Method: Extended interval Newton's method.                                 }
{ Global procedures and functions:                                           }
{    procedure AllZeros(...)      : computes enclosures for all zeros        }
{    function  AllZerosErrMsg(..) : delivers an error message text           }
{ Updates:                                                                   }
{    23.11.1995 : Termination criterium changed to ' <= Epsilon'             }
{----------------------------------------------------------------------------}
module nlfzero;

use
  b_util,    { Boolean utilities            }
  i_ari,     { Interval arithmetic          }
  i_util,    { Interval utilities           }
  xi_ari,    { Extended interval arithmetic }
  ddf_ari;   { Differentiation arithmetic   }

{----------------------------------------------------------------------------}
{ Error codes used in this module.                                           }
{----------------------------------------------------------------------------}
const
  NoError       = 0;  { No error occurred.                                   }
  lbZeroVecNot1 = 1;  { Lower bound of variable ZeroVector is not equal to 1.}
  lbInfoVecNot1 = 2;  { Lower bound of variable InfoVector is not equal to 1.}
  VecsDiffer    = 3;  { Bounds of ZeroVector and InfoVector do not match.    }
  VecTooSmall   = 4;  { ZeroVector too small. Not all zeros can be stored.   }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function AllZerosErrMsg ( Err : integer ) : string;
var
  Msg : string;
begin
  case Err of
    NoError       : Msg := '';
    lbZeroVecNot1 : Msg := 'Lower bound of ZeroVector is not equal to 1';
    lbInfoVecNot1 : Msg := 'Lower bound of InfoVector is not equal to 1';
    VecsDiffer    : Msg := 'Bounds of ZeroVector and InfoVector do not match';
    VecTooSmall   : Msg := 'Not all zeros found. ZeroVector is too small';
    else          : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  AllZerosErrMsg := Msg;
end;

{----------------------------------------------------------------------------}
{ Purpose: Recursive procedure for the execution of the extended interval    }
{    Newton's method for the function 'f'.                                   }
{ Parameters:                                                                }
{    In     : 'f'          : must be declared for the 'DerivType' to enable  }
{                            the internal use of the differentiation         }
{                            arithmetic 'ddf_ari'.                           }
{             'y'          : specifies the starting interval.                }
{             'Epsilon'    : specifies the desired relative accuracy         }
{                            (interval diameter) of the result intervals.    }
{             'yUnique'    : signals whether it is already verified that the }
{                            actual interval 'y' contains a unique zero.     }
{    Out    : 'ZeroVector' : stores the enclosures for the zeros of 'f'.     }
{             'InfoVector' : stores the corresponding information on the     }
{                            uniqueness of the zero in each enclosure.       }
{    In/Out : 'ZeroNo'     : represents the number of the zero computed last }
{                            (in) and the total number of enclosures         }
{                            computed at the end of the recursion (out).     }
{ Description:                                                               }
{    The procedure 'XINewton' is recursively called whenever the extended    }
{    interval division results in a bisection of the actual interval 'y' in  }
{    two intervals 'y1' and 'y2', and the tolerance condition is not ful-    }
{    filled yet. Otherwise the enclosures for the zeros of 'f' are stored in }
{    the interval vector 'ZeroVector', the corresponding information on the  }
{    uniqueness of the zero in each enclosure is stored in the Boolean       }
{    vector 'InfoVector'.                                                    }
{----------------------------------------------------------------------------}
procedure XINewton (function f (x: DerivType) : DerivType;
                                            y : interval;
                                      Epsilon : real;
                                      yUnique : boolean;
                         var       ZeroVector : ivector;
                         var       InfoVector : bvector;
                         var           ZeroNo : integer);
var
  z           : xinterval;
  fc, fy, dfy : interval;
  yp          : ivector[1..2];
  c           : real;
  i           : integer;
begin
  dfEval(f, y, fy, dfy);               { Compute f(y) and f'(y)              }
  if 0 in fy then                      { Start if 0 in f(y), else do nothing }
  begin                                {-------------------------------------}
    c := mid(y);
    fEval(f, intval(c), fc);  { Compute f(c) and f'(c).                      }
    z := c - fc div dfy;      { Extended interval Newton step.               }
    yp:= y ** z;              { Intersect interval y and extended interval z }
                              { resulting in two intervals yp[1] and yp[2].  }
                              {----------------------------------------------}
    if yp[1] = y then                                   { Stagnation, so 'y' }
    begin                                               { must be bisected.  }
      yp[1]:= intval(y.inf,c);  yp[2]:= intval(c,y.sup);{--------------------}
    end;

    if (yp[1] <> EmptyIntval) and (yp[2] = EmptyIntval) then
      yUnique := yUnique or (yp[1] in y)   { Inner inclusion ===> uniqueness }
    else                                   {---------------------------------}
      yUnique := false;

    for i:=1 to 2 do
      if yp[i] <> EmptyIntval then
      begin
        if (RelDiam(yp[i]) <= Epsilon) then
          begin { No more Newton steps }
            fEval(f, yp[i], fy);                 { Compute f(yp[i])          }
            if (0 in fy) then                    { Store enclosure and info  }
            begin                                {---------------------------}
              ZeroNo := ZeroNo + 1;
              if (ZeroNo <= ub(ZeroVector)) then
              begin
                ZeroVector[ZeroNo] := yp[i];       { Store enclosure of zero }
                InfoVector[ZeroNo] := yUnique;     { Store uniqueness info   }
              end;                                 {-------------------------}
            end;
          end
        else { Recursive call of 'XINewton' for interval 'yp[i]' }
         XINewton(f,yp[i],Epsilon,yUnique,ZeroVector,InfoVector,ZeroNo);
      end;
  end;
end;

{----------------------------------------------------------------------------}
{ Purpose: Execution of a verification step including the use of an epsilon  }
{    inflation.                                                              }
{ Parameters:                                                                }
{    In     : 'f'       : function of 'DerivType'.                           }
{    Out    : 'yUnique' : returns 'true' if the verification is successful.  }
{    In/Out : 'y'       : interval enclosure to be verified.                 }
{ Description: This procedure checks the uniqueness of the zero enclosed in  }
{    the variable 'y' by an additional verification step including the use   }
{    of an epsilon inflation of the iterates.                                }
{----------------------------------------------------------------------------}
procedure VerificationStep (function f (x: DerivType) : DerivType;
                                 var                y : interval;
                                 var          yUnique : boolean);
const
  kmax = 10;
var
  yIn, yOld, fc, fy, dfy : interval;
  c, eps                 : real;
  k                      : integer;
begin
  k := 0;  yIn := y;  eps:= 0.25;       { Initializations                    }
  yUnique := false;                     {------------------------------------}
  while (not yUnique) and (k < kmax) do { Do kmax loops to achieve inclusion }
  begin                                 {------------------------------------}
    yOld := blow(y, eps);               { Epsilon inflation of 'y'           }
    dfEval(f, yOld, fy, dfy);           {------------------------------------}

    if 0 in dfy then
      k := kmax                                   { No verification possible }
    else                                          {--------------------------}
      begin                                   { Perform interval Newton step }
        k := k+1;  c := mid(yOld);            {------------------------------}
        fEval(f,intval(c),fc);
        y := c - fc / dfy;

        if y >< yOld then                         { No verification possible }
          k := kmax                               {--------------------------}
        else
          begin
            yUnique := y in yOld;          { Inner inclusion ===> uniqueness }
            y       := y ** yOld;          { Intersection with old value     }
            if y = yOld then               {---------------------------------}
              eps := eps * 8               { Increase the value of 'eps'.    }
          end;                             {---------------------------------}
      end;
  end;
  if not yUnique then y := yIn;
end;

{----------------------------------------------------------------------------}
{ Purpose: Computation of enclosures for all zeros of a continuously         }
{    differentiable one-dimensional, real-valued function.                   }
{ Parameters:                                                                }
{    In     : 'f'             : objective function, must be declared for the }
{                               'DerivType' to enable the internal use of    }
{                               the differentiation arithmetic 'ddf_ari'.    }
{             'Start',        : specifies the starting interval.             }
{             'Epsilon'.      : specifies the desired relative accuracy      }
{                               (interval diameter) of the result intervals. }
{    Out    : 'ZeroVector'    : stores and returns the enclosures for the    }
{                               zeros of 'f'.                                }
{             'InfoVector'    : stores the corresponding information on the  }
{                               uniqueness of the zeros in these enclosures. }
{             'NumberOfZeros' : returns the number of enclosures computed.   }
{             'Err'           : returns an error code.                       }
{ Description:                                                               }
{    The enclosures for the zeros of 'f' are computed by calling procedure   }
{    'XINewton'. Then an additional verification step is applied to those    }
{    enclosures which have not been verified.                                }
{    If an error occurs the value of 'Err' is different from 0.              }
{----------------------------------------------------------------------------}
global procedure AllZeros (function f (x: DerivType) : DerivType;
                                               Start : interval;
                                             Epsilon : real;
                                var       ZeroVector : ivector;
                                var       InfoVector : bvector;
                                var    NumberOfZeros : integer;
                                var              Err : integer);
var
  i          : integer;
  MinEpsilon : real;
begin
  NumberOfZeros:= 0;
  if (lb(ZeroVector) <> 1) then       { Check index bounds of result vectors }
    Err := lbZeroVecNot1              {--------------------------------------}
  else if (lb(InfoVector) <> 1) then
    Err := lbInfoVecNot1
  else if (ub(InfoVector) <> ub(ZeroVector)) then
    Err := VecsDiffer
  else
    begin                            { Start extended interval Newton method }
      Err := NoError;                {---------------------------------------}
      MinEpsilon:= succ(1.0) - 1.0;   { Relative machine accuracy (1 ulp)    }
                                      {--------------------------------------}
      if (Epsilon < MinEpsilon) then Epsilon := MinEpsilon;  { Set 'Epsilon' }
                                                             { to 1 ulp acc. }
      XINewton(f, Start, Epsilon, false,                     {---------------}
                  ZeroVector, InfoVector, NumberOfZeros);

                                             { Check if there are more zeros }
      if ub(ZeroVector) < NumberOfZeros then { than storage space            }
      begin                                  {-------------------------------}
        Err:= VecTooSmall;  NumberOfZeros:= ub(ZeroVector);
      end;
                                                        { Verification step  }
      for i:=1 to NumberOfZeros do                      { for the enclosures }
        if InfoVector[i] <> true then                   {--------------------}
          VerificationStep(f,ZeroVector[i],InfoVector[i]);
    end;
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
