{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Evaluation of a real polynomial p with maximum accuracy.          }
{ Method: Transformation of Horner's scheme for evaluating p(t) to a linear  }
{    system of equations A(t)*x = p, where A(t) is a bidiagonal, Toeplitz    }
{    matrix  and x = (x_n,...,x_0). By iterative refinement the floating-    }
{    point approximation is improved and the exact solution x is enclosed in }
{    an interval vector [x].                                                 }
{ Global procedures and functions:                                           }
{    procedure RPolyEval(...)      : computes an enclosure for the exact     }
{                                    value p(t) with maximum accuracy        }
{    function RPolyEvalErrMsg(...) : delivers an error message text          }
{----------------------------------------------------------------------------}
module rpeval; { Real polynomial evaluation        }
               { -    -          ----              }
use
  rpoly,       { Real polynomials                  }
  i_ari,       { Interval arithmetic               }
  i_util,      { Utilities for type real           }
  mv_ari,      { Matrix/Vector arithmetic          }
  mvi_ari;     { Matrix/Vector interval arithmetic }

const
  kmax = 10;   { Maximum number of iteration steps }
{----------------------------------------------------------------------------}
{ Error codes used in this module. In the comments below p[0],..., p[n] are  }
{ the coefficients of a polynomial p.                                        }
{----------------------------------------------------------------------------}
const
  NoError   = 0;    { No error occurred                             }
  ItFailed  = 1;    { Maximum number of iterations exceeded         }
  IllProb   = 2;    { Illegal Problem, i.e. polynomials with lower  }
                    { bound <> 0 was passed to procedure RPolyEval. }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function RPolyEvalErrMsg(Err : integer) : string;
var
  Msg : string;
begin
  case Err of
    NoError : Msg := '';
    ItFailed: Msg := 'Maximum number of iterations (=' + image(kmax,0) +
                      ') exceeded';
    IllProb : Msg := 'Illegal polynomial with lower bound <> 0 occurred';
    else    : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  RPolyEvalErrMsg := Msg;
end;

{----------------------------------------------------------------------------}
{ Purpose: Determination of p(t) (a polynomial p with argument t) with       }
{    maximum accuracy.                                                       }
{ Parameters:                                                                }
{    In    : 'p'  : represents a real polynomial by its coefficients         }
{            't'  : specifies the point of evaluation                        }
{    Out   : 'z'  : floating-point approximation computed by Horner's scheme }
{            'zz' : enclosure of p(t)                                        }
{            'k'  : number of iterations needed                              }
{            'Err': error flag                                               }
{ Description: The polynomial 'p' is defined by its coefficients and, 't'    }
{    denotes the evaluation point. Horner's scheme for evaluating p is equi- }
{    valent to computing the solution of a linear system of equations        }
{    A(t)*x = p, where A(t) is a bidiagonal, Toeplitz matrix and             }
{    x =(x_n,...,x_0). The component x_0 of x is equal to p(t). The solution }
{    x is enclosed in an interval vector [x] by iterative refinement. The    }
{    first element of [x] is an enclosure of p(t). It is returned in the     }
{    variable 'zz'. A floating-point approximation computed by Horner's      }
{    scheme is returned in 'z'. The number of iterations is returned in 'k'  }
{    and the state of success is returned in 'Err'.                          }
{ Remark: The polynomial's coefficients and the argument are assumed to be   }
{    exact floating-point numbers!                                           }
{----------------------------------------------------------------------------}
global procedure RPolyEval (    p      : RPolynomial;
                                t      : real;
                            var z      : real;
                            var zz     : interval;
                            var k, Err : integer     );
var
  i, j, n : integer;                        { The j-th column of the matrix  }
  rr, yy  : ivector[0..ub(p)];              { 'y' is used to store the j-th  }
  x       : rvector[0..ub(p)];              { correction for the solution of }
  y       : rmatrix[0..ub(p),0..kmax];      { A(t)*y = p.                    }
begin                                       {--------------------------------}
  Err := NoError; n := ub(p); k := 0;                       { Initialization }
                                                            {----------------}
  if (lb(p) <> 0) then                      { Illegal polynomial declaration }
    Err := IllProb                          {--------------------------------}
  else { lb(p) = 0 }
    begin
      if (n = 0) then
        begin  z := p[0];  zz := p[0];  end
      else if (n = 1) then
        begin
          z  := #*( t*p[1] + p[0] );
          zz := ##( t*p[1] + p[0] );
        end
      else { n > 1 }
        begin
          x := Null(x); y := Null(y);     { Initialization x := 0 and y := 0 }
                                          {----------------------------------}
          x[n] := p[n];               { Computation of a first approximation }
          for i := n-1 downto 0 do    { using Horner's scheme                }
            x[i] := x[i+1]*t+p[i];    {--------------------------------------}
          z := x[0];

          y[*,0] := x;                        { Iterative refinement for the }
          repeat                              { solution of A*x = p          }
                                              {------------------------------}
            if (k > 0) then      { If a residual was computed, its middle is }
              y[*,k] := mid(yy); { stored as the next correction of 'y'.     }
                                 {-------------------------------------------}
            yy[n] := 0;                { Computation of the residual [r] and }
            for i := n-1 downto 0 do   { evaluation of the interval system   }
            begin                      { A*[y] = [r]                         }
                                       {-------------------------------------}
              rr[i] := ##( p[i] - for j := 0 to k sum( y[i,j] )
                                + for j := 0 to k sum( t*y[i+1,j] ) );
              yy[i] := yy[i+1]*t + rr[i]
            end;

                              { Determination of a new enclosure [z] of p(t) }
            zz := ##( for j := 0 to k sum( y[0,j] ) + yy[0] );
            k := k + 1;
          until UlpAcc(zz,1) or (k > kmax);
          if not UlpAcc(zz,1) then Err := ItFailed;
        end; { n > 1 }
    end; { lb(p) = 0 }
end; { procedure RPolyEval }

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
